@@ -35,7 +35,6 @@ import subprocess
 import sys
 
 import molecule
-import molecule.config
 import molecule.util
 
 try:
@@ -136,6 +135,11 @@ options:
     required: True
     choices: ['up', 'halt', 'destroy']
     default: None
+  workdir:
+    description:
+      - vagrant working directory
+    required: False
+    default: content of MOLECULE_EPHEMERAL_DIRECTORY environment variable
 requirements:
     - python >= 2.6
     - python-vagrant
@@ -147,214 +151,119 @@ See doc/source/configuration.rst
 """
 
 VAGRANTFILE_TEMPLATE = """
-require 'yaml'
+{%- macro ruby_format(value) -%}
+  {%- if value is boolean -%}
+    {{ value | string | lower }}
+  {%- elif value is string -%}
+    {# "Bug" compat. To be removed later #}
+    {%- if value[0] == value[-1] and value.startswith(("'", '"')) -%}
+    {{ value }}
+    {%- else -%}
+    "{{ value }}"
+    {%- endif -%}
+  {%- else -%}
+    {{ value }}
+  {%- endif -%}
+{%- endmacro -%}
+
+{%- macro dict2args(dictionary) -%}
+  {% set sep = joiner(", ") %}
+  {%- for key, value in dictionary.items() -%}
+    {{ sep() }}{{ key }}: {{ ruby_format(value) }}
+  {%- endfor -%}
+{%- endmacro -%}
 
 Vagrant.configure('2') do |config|
-  vagrant_config_hash = YAML::load_file('{{ vagrantfile_config }}')
-
   if Vagrant.has_plugin?('vagrant-cachier')
+    {% if instance.config_options['cachier'] is sameas false %}
+    config.cache.disable!
+    {% else %}
     config.cache.scope = 'machine'
+    {% endif %}
   end
 
-  ##
-  # Configs
-  ##
+  config.vm.define "{{ instance.name }}" do |c|
+    ##
+    # Box definition
+    ##
+    c.vm.box = "{{ instance.box }}"
+    {{ 'c.vm.box_version = "{}"'.format(instance.box_version) if instance.box_version }}
+    {{ 'c.vm.box_url = "{}"'.format(instance.box_url) if instance.box_url }}
 
-  c = vagrant_config_hash['config']
-  if !c['options']['synced_folder']
-    config.vm.synced_folder ".", "/vagrant", disabled: true
-  end
-  c['options'].delete('synced_folder')
+    ##
+    # Config options
+    ##
+    {% if instance.config_options['synced_folder'] is sameas false %}
+    c.vm.synced_folder ".", "/vagrant", disabled: true
+    {% endif %}
 
-  c['options'].each { |key, value|
-    eval("config.#{key} = #{value}")
-  }
+    {% for k,v in instance.config_options.items() %}
+    {% if k not in ['synced_folder', 'cachier'] %}c.{{ k }} = {{ ruby_format(v) }}{% endif %}
+    {% endfor %}
 
-  ##
-  # Platforms
-  ##
+    c.vm.hostname = "{{ instance.name }}"
 
-  platform = vagrant_config_hash['platform']
+    ##
+    # Network
+    ##
+    {% for n in instance.networks %}
+    c.vm.network "{{ network.name }}", {{ dict2args(network.options) }}
+    {% endfor %}
 
-  config.vm.box = platform['box']
+    ##
+    # instance_raw_config_args
+    ##
+    {% if instance.instance_raw_config_args is not none %}{% for arg in instance.instance_raw_config_args -%}
+    c.{{ arg }}
+    {% endfor %}{% endif %}
 
-  if platform['box_version']
-    config.vm.box_version = platform['box_version']
-  end
+    ##
+    # Provider
+    ##
+    c.vm.provider "{{ instance.provider }}" do |{{ instance.provider | lower }}, override|
+      {% if instance.provider.startswith('vmware_') %}
+      {{ instance.provider | lower }}.vmx['memsize'] = {{ instance.memory }}
+      {{ instance.provider | lower }}.vmx['numvcpus'] = {{ instance.cpus }}
+      {% else %}
+      {{ instance.provider | lower }}.memory = {{ instance.memory }}
+      {{ instance.provider | lower }}.cpus = {{ instance.cpus }}
+      {% endif %}
 
-  if platform['box_url']
-    config.vm.box_url = platform['box_url']
-  end
+      {% for option, value in instance.provider_options.items() %}
+      {{ instance.provider | lower }}.{{ option }} = {{ ruby_format(value) }}
+      {% endfor %}
 
-  ##
-  # Provider
-  ##
+      {% if instance.provider_raw_config_args is not none %}
+        {% for arg in instance.provider_raw_config_args %}
+      {{ instance.provider | lower }}.{{ arg }}
+        {% endfor %}
+      {% endif %}
 
-  provider = vagrant_config_hash['provider']
-  provider_memory = provider['options']['memory']
-  provider_cpus = provider['options']['cpus']
-  provider['options'].delete('memory')
-  provider['options'].delete('cpus')
+      {% if instance.provider_override_args is not none %}
+        {% for arg in instance.provider_override_args -%}
+      override.{{ arg }}
+        {% endfor %}
+      {% endif %}
 
-  ##
-  # Virtualbox
-  ##
-
-  if provider['name'] == 'virtualbox'
-    config.vm.provider provider['name'] do |virtualbox, override|
-      virtualbox.memory = provider_memory
-      virtualbox.cpus = provider_cpus
-
-      if provider['options']['linked_clone']
-        if Gem::Version.new(Vagrant::VERSION) >= Gem::Version.new('1.8.0')
-          virtualbox.linked_clone = provider['options']['linked_clone']
-        end
-      else
-        if Gem::Version.new(Vagrant::VERSION) >= Gem::Version.new('1.8.0')
-          virtualbox.linked_clone = true
-        end
-      end
-
-      # Custom
-      provider['options'].each { |key, value|
-        if key != 'linked_clone'
-          eval("virtualbox.#{key} = #{value}")
-        end
-      }
-
-      # Raw Configuration
-      if provider['raw_config_args']
-        provider['raw_config_args'].each { |raw_config_arg|
-          eval("virtualbox.#{raw_config_arg}")
-        }
-      end
-
-      if provider['override_args']
-        provider['override_args'].each { |override_arg|
-          eval("override.#{override_arg}")
-        }
-      end
-    end
-
-    # The vagrant-vbguest plugin attempts to update packages
-    # before a RHEL based VM is registered.
-    # TODO: Port from the old .j2, should be done in raw config
-    if (vagrant_config_hash['platform'] =~ /rhel/i) != nil
-      if Vagrant.has_plugin?('vagrant-vbguest')
-        config.vbguest.auto_update = false
-      end
-    end
-  end
-
-  ##
-  # VMware (vmware_fusion, vmware_workstation and vmware_desktop)
-  ##
-
-  if provider['name'].start_with?('vmware_')
-    config.vm.provider provider['name'] do |vmware, override|
-      vmware.vmx['memsize'] = provider_memory
-      vmware.vmx['numvcpus'] = provider_cpus
-
-      # Custom
-      provider['options'].each { |key, value|
-        eval("vmware.#{key} = #{value}")
-      }
-
-      # Raw Configuration
-      if provider['raw_config_args']
-        provider['raw_config_args'].each { |raw_config_arg|
-          eval("vmware.#{raw_config_arg}")
-        }
-      end
-
-      if provider['override_args']
-        provider['override_args'].each { |override_arg|
-          eval("override.#{override_arg}")
-        }
-      end
-    end
-  end
-
-  ##
-  # Parallels
-  ##
-
-  if provider['name'] == 'parallels'
-    config.vm.provider provider['name'] do |parallels, override|
-      parallels.memory = provider_memory
-      parallels.cpus = provider_cpus
-
-      # Custom
-      provider['options'].each { |key, value|
-        eval("parallels.#{key} = #{value}")
-      }
-
-      # Raw Configuration
-      if provider['raw_config_args']
-        provider['raw_config_args'].each { |raw_config_arg|
-          eval("parallels.#{raw_config_arg}")
-        }
-      end
-
-      if provider['override_args']
-        provider['override_args'].each { |override_arg|
-          eval("override.#{override_arg}")
-        }
-      end
-    end
-  end
-
-  ##
-  # Libvirt
-  ##
-
-  if provider['name'] == 'libvirt'
-    config.vm.provider provider['name'] do |libvirt, override|
-      libvirt.memory = provider_memory
-      libvirt.cpus = provider_cpus
-
-      # Custom
-      provider['options'].each { |key, value|
-        eval("libvirt.#{key} = #{value}")
-      }
-
-      # Raw Configuration
-      if provider['raw_config_args']
-        provider['raw_config_args'].each { |raw_config_arg|
-          eval("libvirt.#{raw_config_arg}")
-        }
-      end
-
-      if provider['override_args']
-        provider['override_args'].each { |override_arg|
-          eval("override.#{override_arg}")
-        }
-      end
-    end
-  end
-
-
-  ##
-  # Instances
-  ##
-
-  if vagrant_config_hash['instance']
-    instance = vagrant_config_hash['instance']
-    config.vm.define instance['name'] do |c|
-      c.vm.hostname = instance['name']
-
-      if instance['interfaces']
-        instance['interfaces'].each { |interface|
-          c.vm.network "#{interface['network_name']}",
-                       Hash[interface.select{|k| k != 'network_name'}.map{|k,v| [k.to_sym, v]}]
-        }
-      end
-
-      if instance['raw_config_args']
-        instance['raw_config_args'].each { |raw_config_arg|
-          eval("c.#{raw_config_arg}")
-        }
-      end
+      {% if instance.provider == 'virtualbox' %}
+      {% if 'linked_clone' not in instance.provider_options %}
+      virtualbox.linked_clone = true
+      {% endif %}
+      {% endif %}
+      {% if instance.provider == 'libvirt' %}
+        {% if 'driver' in instance.provider_options and 'qemu' in instance.provider_options['driver'] %}
+          {% if 'cpu_mode' not in instance.provider_options %}
+      # When using qemu instead of kvm, some libvirt systems
+      # will use EPYC as vCPU model inside the new VM.
+      # This will lead to process hang with ssh-keygen -A on alpine.
+      # Not sure were the bug is (libvirt or qemu or alpine or all of them) but using QEMU64
+      # cpu model will work around this. Hopefully, by checking 'cpu_mode' option, it will
+      # allow people to override the model to use.
+      libvirt.cpu_mode = 'custom'
+      libvirt.cpu_model = 'qemu64'
+          {% endif %}
+        {% endif %}
+      {% endif %}
     end
   end
 end
@@ -385,7 +294,7 @@ class VagrantClient(object):
         self._module = module
 
         self._config = self._get_config()
-        self._vagrantfile = self._config.driver.vagrantfile
+        self._vagrantfile = self._config["vagrantfile"]
         self._vagrant = self._get_vagrant()
         self._write_configs()
         self._has_error = None
@@ -394,7 +303,7 @@ class VagrantClient(object):
 
     @contextlib.contextmanager
     def stdout_cm(self):
-        """ Redirect the stdout to a log file. """
+        """Redirect the stdout to a log file."""
         with open(self._get_stdout_log(), "a+") as fh:
             msg = "### {} ###\n".format(self._datetime)
             fh.write(msg)
@@ -404,7 +313,7 @@ class VagrantClient(object):
 
     @contextlib.contextmanager
     def stderr_cm(self):
-        """ Redirect the stderr to a log file. """
+        """Redirect the stderr to a log file."""
         with open(self._get_stderr_log(), "a+") as fh:
             msg = "### {} ###\n".format(self._datetime)
             fh.write(msg)
@@ -444,17 +353,17 @@ class VagrantClient(object):
                 # passes the actual error as a no-argument ContextManager.
                 pass
 
-            # NOTE(retr0h): Ansible wants only one module return `fail_json`
-            # or `exit_json`.
-            if not self._has_error:
-                self._module.exit_json(
-                    changed=changed, log=self._get_stdout_log(), **self._conf()
-                )
-            else:
-                msg = "ERROR: See log file '{}'".format(self._get_stderr_log())
-                with io.open(self._get_stderr_log(), "r", encoding="utf-8") as f:
-                    self.result["stderr"] = f.read()
-                self._module.fail_json(msg=msg, **self.result)
+        # NOTE(retr0h): Ansible wants only one module return `fail_json`
+        # or `exit_json`.
+        if not self._has_error:
+            self._module.exit_json(
+                changed=changed, log=self._get_stdout_log(), **self._conf()
+            )
+
+        msg = "Failed to start the VM: See log file '{}'".format(self._get_stderr_log())
+        with io.open(self._get_stderr_log(), "r", encoding="utf-8") as f:
+            self.result["stderr"] = f.read()
+        self._module.fail_json(msg=msg, **self.result)
 
     def destroy(self):
         changed = False
@@ -477,7 +386,15 @@ class VagrantClient(object):
     def _conf(self):
         instance_name = self._module.params["instance_name"]
 
-        return self._vagrant.conf(vm_name=instance_name)
+        try:
+            return self._vagrant.conf(vm_name=instance_name)
+        except Exception:
+            msg = "Failed to get vagrant config for {}: See log file '{}'".format(
+                instance_name, self._get_stderr_log()
+            )
+            with io.open(self._get_stderr_log(), "r", encoding="utf-8") as f:
+                self.result["stderr"] = f.read()
+                self._module.fail_json(msg=msg, **self.result)
 
     def _status(self):
         instance_name = self._module.params["instance_name"]
@@ -485,10 +402,13 @@ class VagrantClient(object):
             s = self._vagrant.status(vm_name=instance_name)[0]
 
             return {"name": s.name, "state": s.state, "provider": s.provider}
-        except AttributeError:
-            pass
-        except subprocess.CalledProcessError:
-            pass
+        except Exception:
+            msg = "Failed to get status for {}: See log file '{}'".format(
+                instance_name, self._get_stderr_log()
+            )
+            with io.open(self._get_stderr_log(), "r", encoding="utf-8") as f:
+                self.result["stderr"] = f.read()
+                self._module.fail_json(msg=msg, **self.result)
 
     def _created(self):
         status = self._status()
@@ -497,75 +417,83 @@ class VagrantClient(object):
         return {}
 
     def _get_config(self):
-        molecule_file = os.environ["MOLECULE_FILE"]
-
-        return molecule.config.Config(molecule_file)
+        conf = dict()
+        conf["workdir"] = os.getenv("MOLECULE_EPHEMERAL_DIRECTORY")
+        if self._module.params["workdir"] is not None:
+            conf["workdir"] = self._module.params["workdir"]
+        conf["vagrantfile"] = os.path.join(conf["workdir"], "Vagrantfile")
+        return conf
 
     def _write_vagrantfile(self):
         template = molecule.util.render_template(
-            VAGRANTFILE_TEMPLATE,
-            vagrantfile_config=self._config.driver.vagrantfile_config,
+            VAGRANTFILE_TEMPLATE, instance=self._get_vagrant_config_dict()
         )
         molecule.util.write_file(self._vagrantfile, template)
 
-    def _write_vagrantfile_config(self, data):
-        molecule.util.write_file(
-            self._config.driver.vagrantfile_config, molecule.util.safe_dump(data)
-        )
-
     def _write_configs(self):
-        self._write_vagrantfile_config(self._get_vagrant_config_dict())
         self._write_vagrantfile()
+        valid = subprocess.run(
+            ["vagrant", "validate"],
+            cwd=self._config["workdir"],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+        )
+        if valid.returncode != 0:
+            self._module.fail_json(
+                msg="Failed to validate generated Vagrantfile: {}".format(valid.stderr)
+            )
 
     def _get_vagrant(self):
-        env = os.environ.copy()
-        env["VAGRANT_CWD"] = os.environ["MOLECULE_EPHEMERAL_DIRECTORY"]
-        v = vagrant.Vagrant(out_cm=self.stdout_cm, err_cm=self.stderr_cm, env=env)
+        v = vagrant.Vagrant(
+            out_cm=self.stdout_cm,
+            err_cm=self.stderr_cm,
+            root=self._config["workdir"],
+        )
 
         return v
 
     def _get_vagrant_config_dict(self):
+        networks = []
+        for iface in self._module.params["instance_interfaces"]:
+            net = dict()
+            net["name"] = iface["network_name"]
+            iface.pop("network_name")
+            net["options"] = iface
+            networks.append(net)
+
         d = {
-            "config": {
-                # NOTE(retr0h): Options provided here will be passed to
-                # Vagrant as "config.#{key} = #{value}".
-                "options": {
-                    # NOTE(retr0h): `synced_folder` does not represent the
-                    # actual key used by Vagrant.  Is used as a flag to
-                    # simply enable/disable shared folder.
-                    "synced_folder": False,
-                    "ssh.insert_key": True,
-                }
-            },
-            "platform": {
-                "box": self._module.params["platform_box"],
-                "box_version": self._module.params["platform_box_version"],
-                "box_url": self._module.params["platform_box_url"],
-            },
-            "instance": {
-                "name": self._module.params["instance_name"],
-                "interfaces": self._module.params["instance_interfaces"],
-                "raw_config_args": self._module.params["instance_raw_config_args"],
-            },
-            "provider": {
-                "name": self._module.params["provider_name"],
-                # NOTE(retr0h): Options provided here will be passed to
-                # Vagrant as "$provider_name.#{key} = #{value}".
-                "options": {
-                    "memory": self._module.params["provider_memory"],
-                    "cpus": self._module.params["provider_cpus"],
-                },
-                "raw_config_args": self._module.params["provider_raw_config_args"],
-                "override_args": self._module.params["provider_override_args"],
+            "name": self._module.params["instance_name"],
+            "memory": self._module.params["provider_memory"],
+            "cpus": self._module.params["provider_cpus"],
+            "networks": networks,
+            "instance_raw_config_args": self._module.params["instance_raw_config_args"],
+            "config_options": {
+                # NOTE(retr0h): `synced_folder` does not represent the
+                # actual key used by Vagrant.  Is used as a flag to
+                # simply enable/disable shared folder.
+                "synced_folder": False,
+                "ssh.insert_key": True,
+                "cachier": True,
             },
+            "box": self._module.params["platform_box"],
+            "box_version": self._module.params["platform_box_version"],
+            "box_url": self._module.params["platform_box_url"],
+            "provider": self._module.params["provider_name"],
+            "provider_options": {},
+            "provider_raw_config_args": self._module.params["provider_raw_config_args"],
+            "provider_override_args": self._module.params["provider_override_args"],
         }
 
-        molecule.util.merge_dicts(
-            d["config"]["options"], self._module.params["config_options"]
+        d["config_options"].update(
+            molecule.util.merge_dicts(
+                d["config_options"], self._module.params["config_options"]
+            )
         )
 
-        molecule.util.merge_dicts(
-            d["provider"]["options"], self._module.params["provider_options"]
+        d["provider_options"].update(
+            molecule.util.merge_dicts(
+                d["provider_options"], self._module.params["provider_options"]
+            )
         )
 
         return d
@@ -580,8 +508,7 @@ class VagrantClient(object):
         instance_name = self._module.params["instance_name"]
 
         return os.path.join(
-            self._config.scenario.ephemeral_directory,
-            "vagrant-{}.{}".format(instance_name, __type),
+            self._config["workdir"], "vagrant-{}.{}".format(instance_name, __type)
         )
 
 
@@ -604,6 +531,7 @@ def main():
             provision=dict(type="bool", default=False),
             force_stop=dict(type="bool", default=False),
             state=dict(type="str", default="up", choices=["up", "destroy", "halt"]),
+            workdir=dict(type="str"),
         ),
         supports_check_mode=False,
     )
@@ -619,7 +547,7 @@ def main():
     if module.params["state"] == "halt":
         v.halt()
 
-    module.exit_json(**module.result)
+    module.fail_json(msg="Unknown error", **v.result)
 
 
 if __name__ == "__main__":
